// Utility class for the Poincare disk

class Poincare {
    public static final double SQRT3 = 1.73205080756887729352;
    static final Complex OMEGA = new Complex(-0.5, 0.5 * SQRT3);

    public static double twoAtanh(double x)
    {
	return Math.log((1.0 + x)/(1.0 - x));
    }
    public static double dist(Complex z, Complex w)
    {
	Complex t = z.sub(w).div(z.mul(w.conj()).realAdd(-1.0));
	return twoAtanh(t.norm());
    }
    public static Complex toKlein(Complex z)
    {
	return z.realMul( 2.0 / (1.0 + z.normSq()) );
    }
    public static Complex fromKlein(Complex z)
    {
	return z.realMul( 1.0 / (1.0 + Math.sqrt(1.0 - z.normSq())) );
    }
    public static Complex fromUH(Complex z)
    {
	return z.mul(Complex.I).realAdd(1.0).div(z.add(Complex.I));
    }
    public static double karea(Complex z, Complex w, Complex p)
    {
	Complex Z = toKlein(z);
	Complex W = toKlein(w);
	Complex P = toKlein(p);
	return Geom.signedArea(Z,W,P);
    }
    public static int side(Complex z, Complex w, Complex p)
    {
	Complex Z = toKlein(z);
	Complex W = toKlein(w);
	Complex P = toKlein(p);
	return (int)Geom.side(Z,W,P);
    }
    protected static double heightScale(double t)
    {
	return (SQRT3 / (1.0 - t));
    }
    public static Complex randUpperStrip()
    {
	return new Complex( 2.0 * (Math.random() - 0.5),
			    heightScale(Math.random()) );
    }
    protected static boolean inUpperThird(Complex z)
    {
	return ( (z.realAdd(-1.0).normSq() > 4.0) && 
		 (z.realAdd(1.0).normSq() >= 4.0) );
    }
    public static Complex randomUpperThird()
    {
	Complex z;
	do {
	    z = randUpperStrip();
	} while (!inUpperThird(z));
	return z;
    }
    public static int randomThree()
    {
        double z = Math.random();
	return((int)Math.round(Math.floor(3.0*z)));
    }
    
    public static Complex randomStdIdeal()
    {
	Complex z = fromUH(randomUpperThird().realMul(1.0 / SQRT3)).mul(Complex.I).realMul(-1.0);
	int i = randomThree();
	if (i == 0) {
	    return(z);
	} else if (i == 1) {
	    return z.mul(OMEGA);
	} else {
	    return z.div(OMEGA);
	}
    }
    public static double randomAngle()
    {
	return (Math.PI * 2.0 * Math.random());
    }
    public static Complex randomIdeal(Complex z)
    {
	return (LinFrac.zeroToPoint(z).transform(Complex.cxExp(1.0,randomAngle())));
    }
}
